# -*- coding: utf-8 -*-

import os

# Third Part Imports

# Basics Imports
import configuration

# Utils
from bravaweb.utils import GetHeader

# Security
import bravaweb.security as Security

# Router
from bravaweb.router import Router

# Environment
from bravaweb.environment import Environment

# Error Tratament
import traceback

# App Boot Time
from datetime import datetime

# App Log
from bravaweb.utils.log import *

boot_time = datetime.now()


async def App(scope, receive, send, hanndler_error=None):
    if scope["type"] == "lifespan":
        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                Router.load_routes()
                await send({"type": "lifespan.startup.complete"})
            elif message["type"] == "lifespan.shutdown":
                await send({"type": "lifespan.shutdown.complete"})
                return

    if scope["type"] not in ("http", "websocket"):
        return

    try:
        try:
            Debug(f"Request '{scope.get('method', 'WS')} {scope.get('path', '/')}'")
            headers = GetHeader(scope)
            Debug("Catching Header")
        except Exception as e:
            if not hanndler_error is None:
                hanndler_error(e, scope, None, 'Global Error: Get Header')
            raise e

        try:

            if scope["method"] == "OPTIONS":
                await Security.origin.Options(headers, send)

            elif Security.origin.Static(scope):
                await Security.origin.NotFound(send)

            elif Security.origin.Permitted(headers, scope):

                try:
                    Debug("Load Environment")
                    environment = Environment(headers, scope, receive, send)

                    Debug("Receiving Body")
                    await environment.ReceiveBody()

                    Debug("Start Response")
                    await environment.Response()

                except Exception as e:
                    if not hanndler_error is None:
                        hanndler_error(e, scope, headers, 'Global Error: Environment, Body Response')
                    Error("App Response Content", e)
                    raise e
            else:
                Error("", "Request Origin Not Authorized", False)
                await Security.origin.Forbidden(send)

        except Exception as e:
            if not hanndler_error is None:
                hanndler_error(e, scope, headers, 'Global Error: Unexpected')
            raise e
    except Exception as e:
        # Global fallback if something went wrong outside the normal controller flow
        try:
            # Fallback if environment failed to build or controller failed
            from bravaweb.response.json import Json
            import configuration
            from bravaweb.utils.response_code import ResponseCode
            
            debug = getattr(configuration.api, "debug", False)
            if debug:
                import traceback
                error_details = {
                    "message": str(e) if e else "Internal Server Error",
                    "traceback": traceback.format_exc(),
                    "context": "Global Catch"
                }
                content = Json().Response(error_details, success=False, error="Internal Error")
                _content_type = b'application/json; charset=utf-8'
            else:
                content = "500: Internal Error".encode(configuration.api.encoding)
                _content_type = b'text/html; charset=utf-8'

            _headers = [
                (b'content-type', _content_type),
                (b"Content-Length", str(len(content)).encode(configuration.api.encoding)),
                (b"X-Frame-Options", b"Deny")
            ]

            # Try to add CORS header if environment exists
            if 'environment' in locals() and hasattr(environment, 'origin') and environment.origin:
                 _headers.append((b"Access-Control-Allow-Origin", environment.origin.encode(configuration.api.encoding)))

            await send({
                'type': 'http.response.start',
                'status': ResponseCode.InternalError.code,
                'headers': _headers
            })

            await send({
                'type': 'http.response.body',
                'body': content
            })
        except Exception as secondary_error:
            # Last resort log
            from bravaweb.utils.log import Error
            Error("Fatal error in global exception handler", secondary_error)
            raise e





import datetime
import configuration

from bravaweb.utils import ResponseType
import bravaweb.utils

class ResponseObject():

    def __init__(self):
        self.extension = ResponseType[self.__class__.__name__].extension
        self.content_type = ResponseType[self.__class__.__name__].description
        self.header = ("Content-Type".encode(configuration.api.encoding),
                       self.content_type.encode(configuration.api.encoding))
        try:
            self.template = getattr(bravaweb.utils.templates, self.__class__.__name__)
        except AttributeError:
            self.template = None

    def Response(self, data, **args):
        return data



Xml = "<xml> \
        <token/>\
        <success/>\
        <date/>\
        <items/>\
        <data/>\
       </xml>"

Json = {
    "token": "",
    "success": True,
    "date": "",
    "items": 0,
    "data": [],
}




import json
import datetime
import decimal

# configuration
import configuration

from bravaweb.response.lib.object import ResponseObject
from bravaweb.utils.encoder import CustomEncoder

class Json(ResponseObject):

    def Response(self, data, success=True, **args):
        _data = data

        _template = dict(self.template)
        _template["date"] = datetime.datetime.now()
        _template["items"] = 0
        if success:

            if hasattr(_data, 'toJSON'):
                _data = _data.toJSON()

            _template["data"] = _data
            _template["date"] = datetime.datetime.now()

            try:
                _template["items"] = len(_data) if isinstance(_data, list) else 1
            except TypeError:
                _template["items"] = 1
            except Exception as e:
                _template["items"] = 0

            for k,v in args.items():
                _template[k] = v

        else:
            _template["success"] = False
            _template["data"] = _data
            _template["date"] = datetime.datetime.now()
            _template["error"] = args["error"] if "error" in args else "Invalid Request"

        return json.dumps(_template, cls=CustomEncoder, indent=4).encode(configuration.api.encoding)



# -*- coding: utf-8 -*-


async def GetBody(receive):
    body = b''
    more_body = True
    while more_body:
        message = await receive()
        body += message.get('body', b'')
        more_body = message.get('more_body', False)
    return body



# -*- coding: utf-8 -*-

from bravaweb.utils.request_header import *
from bravaweb.utils.request_body import *
from bravaweb.utils.response_code import *
from bravaweb.utils.response_type import *
from bravaweb.utils.request_methods import *
from bravaweb.utils.templates import *



# -*- coding: utf-8 -*-

import json
import datetime
import decimal

import configuration

class CustomEncoder(json.JSONEncoder):
    def default(self, o):
        try:
            if isinstance(o, decimal.Decimal):
                return float(o)
            if isinstance(o, datetime.datetime):
                return o.strftime(configuration.api.date_format)
            if isinstance(o, bytes):
                return o.decode('utf-8')
            return super(CustomEncoder, self).default(o)
        except Exception as e:
            raise e



# -*- coding: utf-8 -*-

# Third Part
import re
import jwt
import json
import urllib.parse

# configuration
import configuration

# Utils
from bravaweb.utils import GetBody, RequestMethod

# Secutiry
from bravaweb.security.request_params import ValidRequestParams

# App Log
from bravaweb.utils.log import *

from bravaweb.utils.encoder import CustomEncoder

import json

class Environment():

    def __init__(self, headers, scope, receive, send):

        # Init Params
        self.headers = headers
        self.scope = scope
        self.send = send
        self.receive = receive

        # Default Params
        self.origin = headers["origin"] if "origin" in headers else None
        self.remote_ip = scope['client'][0]
        Debug(f"Remote Ip from scope: {self.remote_ip}")

        # GET REAL USER IP IF FROM CDN SERVICE
        trusted_proxies = getattr(configuration.api, "trusted_proxies", [])
        is_trusted_proxy = ('*' in trusted_proxies) or (self.remote_ip in trusted_proxies)

        if is_trusted_proxy:
            if 'cf-connecting-ip' in headers:
                self.remote_ip = headers['cf-connecting-ip']
                Debug(f"Remote Ip from 'CF-Connecting-IP' header: {self.remote_ip}")
            elif 'x-forwarded-for' in headers:
                self.remote_ip = headers['x-forwarded-for'].split(',')[0].strip() if ',' in headers['x-forwarded-for'] else headers['x-forwarded-for']
                Debug(f"Remote Ip from 'X-Forwarded-For' header: {self.remote_ip}")

        self.remote_uuid = headers['uuid'] if "uuid" in headers else None
        self.browser = headers["user-agent"] if "user-agent" in headers else None
        self.accept_encoding = headers["accept-encoding"] if "accept-encoding" in headers else None
        self.remote = {
            "ip": self.remote_ip,
            'uuid' : self.remote_uuid,
            "browser": self.browser,
            "accept": self.accept_encoding
            }
        Debug(f"Origin: {self.origin}")
        Debug(f"Remote: {json.dumps(self.remote, indent=14, sort_keys=True)}")

        self.method = RequestMethod(scope["method"])
        self.response_type = None

        # Authorization
        self.authorization = headers["authorization"] if "authorization" in headers else None

        Debug(f"Authorization: {self.authorization }")

        self.auth_token = None

        try:
            if self.authorization:
                options = {"verify_exp": True} if getattr(configuration.api, "verify_jwt_exp", False) else {}
                self.bearer = jwt.decode(
                    headers["authorization"].replace('Bearer ', ''),
                    configuration.api.token,
                    algorithms=['HS256'],
                    options=options
                )
            else:
                self.bearer = None
        except jwt.ExpiredSignatureError:
            self.bearer = None
            Debug("JWT Authorization Failed: Token Expired")
        except jwt.InvalidTokenError as e:
            self.bearer = None
            Debug(f"JWT Authorization Failed: Invalid Token - {e}")
        except Exception as e:
            self.bearer = None
            Debug(f"JWT Authorization Failed: {e}")

        # Content Length
        self.content_length = int(
            headers['content-length']) if "content-length" in headers else 0

        # Query String
        self.get = {}
        query_string = scope["query_string"].decode(configuration.api.encoding)
        if query_string:
            for item_name, item_value in urllib.parse.parse_qsl(query_string, keep_blank_values=True):
                if "[" in item_name and "]" in item_name:
                    object_name = item_name.split("[")[0]
                    field_name = item_name.split("[")[1].replace("]", "")
                    if object_name not in self.get or not isinstance(self.get[object_name], dict):
                        self.get[object_name] = {}
                    self.get[object_name][field_name] = item_value
                else:
                    if item_name not in self.get or not isinstance(self.get[item_name], dict):
                        self.get[item_name] = item_value

        # Post
        self.post = {}

        # Validate Route
        self.controller, self.area, self.module, self.action, self.id = None, None, "default", "index", None

        # Route Rule
        RouteRule = [route for route in configuration.api.routes
                     if bool(re.match(route[1], scope["path"]))][0][0]

        Debug(f"Route Rule: {RouteRule}")

        # Request Route
        RequestRoute = scope["path"][1:].split("/")
        for index, attr in enumerate(RouteRule.split("/")):
            # If Attribute was defined
            if re.match(r"\{.*?\}", attr):
                try:
                    self.__setattr__(
                        attr[attr.find("{") + 1:attr.find("}")], RequestRoute[index])
                except IndexError:
                    pass
            # Else Pattern
            elif index == 0:
                self.controller = attr
            elif index == 1:
                self.area = attr
            elif index == 2:
                self.module = attr
            elif index == 3:
                self.action = attr
            elif index == 4:
                self.id = attr

        # Validate safe names to prevent Path Traversal
        for key in ['controller', 'area', 'module', 'action']:
            val = getattr(self, key, None)
            if val and not bool(re.match(r"^[a-zA-Z0-9_-]+$", val)):
                setattr(self, key, None)

        self.route = ["controllers"]


        if self.controller:
            self.route.append(self.controller)
        if self.area:
            self.route.append(self.area)
        if self.module:
            self.route.append(self.module)

        Debug(f"Controller: {self.controller}")
        Debug(f"Area: {self.area}")
        Debug(f"Module: {self.module}")
        Debug(f"Action: {self.action}")
        Debug(f"ID: {self.id}")

        self.klass = "{}Controller".format(self.route[-1].capitalize())

    async def ReceiveBody(self):
        self.body = await GetBody(self.receive)

        # Post
        if len(self.body) > 0:
            try:
                self.post = json.loads(
                    self.body.decode(configuration.api.encoding))
            except ValueError:
                body_string = self.body.decode(configuration.api.encoding)
                if body_string:
                    for item_name, item_value in urllib.parse.parse_qsl(body_string, keep_blank_values=True):
                        if "[" in item_name and "]" in item_name:
                            object_name = item_name.split("[")[0]
                            field_name = item_name.split("[")[1].replace("]", "")
                            if object_name not in self.post or not isinstance(self.post[object_name], dict):
                                self.post[object_name] = {}
                            self.post[object_name][field_name] = item_value
                        else:
                            if item_name not in self.post or not isinstance(self.post[item_name], dict):
                                self.post[item_name] = item_value

        # Union Get and Post Params
        self.request = {**self.get, **self.post}

        Debug(f"Request: {json.dumps(self.request, cls=CustomEncoder, indent=14, sort_keys=True)}")

    async def Response(self):

        _class_ref, _class, _action, params = None, None, None, {}

        # Load Class from Router Cache
        try:
            from bravaweb.router import Router
            Debug(f"Catching Class via Router: '{'.'.join(self.route)}' -> '{self.klass}'")
            _class_ref = Router.get_controller_class(self.route, self.klass)
        except Exception as e:
            Error("Load Class via Router", e)
            raise e

        if _class_ref is None:
            Debug(f"Class not found/mapped: '{self.klass}'")
            _action = __import__("bravaweb.controller").__getattribute__("Controller")(self).__getattribute__("NotFound")
        else:
            try:
                try:
                    _class = _class_ref(environment=self)
                except TypeError:
                    try:
                        _class = _class_ref(envirom=self)
                    except TypeError:
                        try:
                            _class = _class_ref(enviroment=self)
                        except TypeError:
                            _class = _class_ref(self)
            except Exception as e:
                Error("Instancing Class", e)
                raise e

        # Load Action
        if _class:
            try:
                Debug(f"Catching Action: 'async def {self.action}(...'")
                _action = _class.__getattribute__(self.action)

                is_safe = False
                _current_action = _action
                _visited_actions = set()
                while _current_action and id(_current_action) not in _visited_actions:
                    _visited_actions.add(id(_current_action))
                    if getattr(_current_action, "is_route", False):
                        is_safe = True
                        break
                    
                    if hasattr(_current_action, "__wrapped__"):
                        _current_action = _current_action.__wrapped__
                        continue
                        
                    if hasattr(_current_action, "__func__"):
                        _current_action = _current_action.__func__
                        continue
                        
                    if hasattr(_current_action, "__annotations__") and "return" in _current_action.__annotations__:
                        _ret = _current_action.__annotations__["return"]
                        if callable(_ret):
                            _current_action = _ret
                            continue
                            
                    break
                    
                if not is_safe:
                    Debug(f"Security: Action '{self.action}' is not marked as a safe route")
                    raise AttributeError("Action is not exposed via route decorators")

                # Validate Request Action
                try:
                    if _action:
                        Debug(f"Validating Required Parameters")
                        params, errors = ValidRequestParams(self, _action, _class)
                        if len(errors) > 0:
                            Debug(f"Validate fail: {errors}")
                            _action = _class.__getattribute__("PreconditionFailed")
                            params = {'errors': errors}
                except Exception as e:
                    Error("Validade Request Params", e)
                    raise e

            except AttributeError as e:
                Error("Attribute Params Error", e)
                # 404 Not Found
                Debug(f"Action not found: 'async def {self.action}(...)")
                _action = __import__("bravaweb.controller").__getattribute__("Controller")(self).__getattribute__("NotFound")
            except Exception as e:
                _action = __import__("bravaweb.controller").__getattribute__("Controller")(self).__getattribute__("InternalError")
                await _action(e)

        # Execute Action
        try:
            Debug(f"Execute Action {self.action}")
            await _action(**params)
        except Exception as e:
            Error("Execute Action", e)
            raise e



# -*- coding: utf-8 -*-

import configuration


def GetHeader(scope):
    headers = {}
    for item in scope["headers"]:
        headers[item[0].decode(configuration.api.encoding)] = item[1].decode(
            configuration.api.encoding)
    return headers




import traceback
import configuration
from datetime import datetime
import inspect

from bravaweb.utils.log import *

def ValidRequestParams(environment, action, klass):

    params = {}
    errors = []
    annotations = action.__annotations__

    # WHILE DECORATOR, GET ENCAPSULATE FUNCTION
    if "return" in annotations:
        try:
            # If it's a decorator, it usually has a function name containing 'decorator'
            # We follow the chain of decorators to find the actual function
            while hasattr(annotations["return"], "__name__") and "decorator" in annotations["return"].__name__:
                if hasattr(annotations["return"], "__annotations__"):
                    annotations = annotations["return"].__annotations__
                else:
                    break

            # If the current 'return' itself has annotations (e.g. it's the wrapped function), use them
            if hasattr(annotations["return"], "__annotations__"):
                annotations = dict(annotations["return"].__annotations__)
        except Exception:
            pass

        # SET ENVIRONMENT RESPONSE TYPE
        if "return" in annotations:
            environment.response_type = annotations["return"]

        # REMOVE RETURN FROM REQUIRED PARAMS
        try:
            del annotations["return"]
        except Exception as e:
            pass

        # GET DEFAULTS
        try:
            sig = inspect.signature(action)
            action_params = sig.parameters
        except ValueError:
            action_params = {}

        # GET REQUIRED PARAM IN REQUEST
        for param, instance in annotations.items():
            has_default = param in action_params and action_params[param].default != inspect.Parameter.empty
            
            if param not in environment.request:
                if not has_default:
                    errors.append(
                        "Field [{}] not present in the request".format(param))
            else:
                try:
                    instance_name = getattr(instance, "__name__", "")
                    if isinstance(environment.request[param], dict):
                        params[param] = instance(**environment.request[param])
                    elif instance_name == "bool" and isinstance(environment.request[param], str):
                        params[param] = environment.request[param].lower() in ['true', '1']
                    elif instance_name == "datetime" and isinstance(environment.request[param], str):
                        params[param] = datetime.strptime(environment.request[param], configuration.api.date_format)
                    else:
                        try:
                            params[param] = instance(environment.request[param])
                        except TypeError:
                            # Se a tipagem falhar (ex: typing.List), preserva o dado bruto
                            params[param] = environment.request[param]
                except ValueError:
                    errors.append("Invalid value for [{}] field".format(param))
                except Exception as e:
                    errors.append("Unexpected [{}] field: {}".format(param, str(e)))

        if len(errors) > 0:
            Error("Parameter Validation", "\n".join(errors), traceback=False)

    return params, errors



# -*- coding: utf-8 -*-

import os
import importlib
import configuration
from bravaweb.utils.log import Debug, Error

class Router:
    _controller_cache = {}

    @classmethod
    def load_routes(cls):
        """
        Percorre o diretório 'controllers' da aplicação no boot e faz o pré-carregamento 
        de todos os módulos e suas classes Controller, armazenando em cache.
        """
        if not hasattr(configuration, 'api') or not hasattr(configuration.api, 'directory'):
            return
            
        controllers_path = os.path.join(configuration.api.directory, "controllers")
        
        if not os.path.exists(controllers_path):
            Debug("O diretório 'controllers' não foi encontrado. Roteamento dinâmico sem pré-cache.")
            return

        Debug("Iniciando pré-carregamento (Boot) das rotas e controladores...")

        for root, dirs, files in os.walk(controllers_path):
            if '__pycache__' in root:
                continue

            for file in files:
                if file.endswith(".py") and file != "__init__.py":
                    rel_dir = os.path.relpath(root, configuration.api.directory)
                    module_name = file[:-3]
                    
                    if rel_dir == ".":
                        module_path = module_name
                    else:
                        module_path = f"{rel_dir.replace(os.sep, '.')}.{module_name}"

                    try:
                        _module = importlib.import_module(module_path)
                        klass = f"{module_name.capitalize()}Controller"
                        
                        if hasattr(_module, klass):
                            _class_ref = getattr(_module, klass)
                            cache_key = f"{module_path}:{klass}"
                            cls._controller_cache[cache_key] = _class_ref
                            Debug(f"Rota em cache pronta: {cache_key}")
                    except Exception as e:
                        if getattr(configuration.api, "debug", False):
                            raise e
                        Error("Erro ao pré-carregar módulo", e)

        Debug(f"Total de classes controller carregadas: {len(cls._controller_cache)}")

    @classmethod
    def get_controller_class(cls, route_list, klass_name):
        """
        Retorna a referência da classe Controller.
        Busca do cache, ou tenta importar dinamicamente caso não esteja no cache.
        """
        module_path = ".".join(route_list)
        cache_key = f"{module_path}:{klass_name}"

        if cache_key in cls._controller_cache:
            return cls._controller_cache[cache_key]

        # Fallback (Lazy Load)
        try:
            _module = importlib.import_module(module_path)
            if hasattr(_module, klass_name):
                _class_ref = getattr(_module, klass_name)
                cls._controller_cache[cache_key] = _class_ref
                return _class_ref
        except Exception as e:
            if getattr(configuration.api, "debug", False):
                raise e
            # Emite log do erro caso o arquivo exista mas contenha erros de sintaxe ou imports inválidos
            Error(f"Erro ao carregar Controller via rota '{module_path}'", e)

        return None



import datetime
import configuration

from bravaweb.utils import ResponseCode

from bravaweb.utils.log import *


async def View(environment, data, **args):

    Debug(f"View: Start Response Type {environment.response_type.__name__}")

    if not "token" in args and environment.auth_token:
        args["token"] = environment.auth_token

    response = environment.response_type()

    content = response.Response(data=data, route=environment.route, action=environment.action, **args)

    _headers = [
        response.header,
        (b"Content-Length", str(len(content)).encode(configuration.api.encoding)),
        (b"Accept-Ranges", b"bytes"),
        (b"X-Frame-Options", b"Deny")
    ]

    if environment.origin:
        _headers.append((b"Access-Control-Allow-Origin", environment.origin.encode(configuration.api.encoding)))

    await environment.send({
        'type': 'http.response.start',
        'status': ResponseCode.OK.code,
        'headers': _headers
    })

    await environment.send({
        'type': 'http.response.body',
        'body': content
    })
    
    Debug(f"View: Finalized Response")




# Decorators Request Methods
from bravaweb.utils import RequestMethod
from bravaweb.utils.log import *

def TestMethod(required, cls, f, **args):
    if required == cls.environment.method:
        return f(cls, **args)
    else:
        Debug(f"Request method not allowed: {cls.environment.method}  - expected {required}")
        return cls.NotAllowed()


def post(f):
    def method_decorator(cls, **args) -> f:
        return TestMethod(RequestMethod.Post, cls, f, **args)
    method_decorator.is_route = True
    return method_decorator


def get(f):
    def method_decorator(cls, **args) -> f:
        return TestMethod(RequestMethod.Get, cls, f, **args)
    method_decorator.is_route = True
    return method_decorator


def put(f):
    def method_decorator(cls, **args) -> f:
        return TestMethod(RequestMethod.Put, cls, f, **args)
    method_decorator.is_route = True
    return method_decorator


def delete(f):
    def method_decorator(cls, **args) -> f:
        return TestMethod(RequestMethod.Delete, cls, f, **args)
    method_decorator.is_route = True
    return method_decorator



#!/usr/bin/env python

# LIBS
from datetime import datetime
import traceback as trace

# LOGGING
import logging

# CREATE TOP LOG LEVEL
log = logging.getLogger("uvicorn.error")

def Info(message):
    log.info(message)

def Debug(message):
    log.debug(message)

def Warning(message):
    log.warning(message)

def Error(sector, e, traceback=True):
    _trace = None
    message = ""
    if traceback:
        _trace = trace.format_exc()
        message = f"\n {sector if sector else ''}: {_trace if _trace else ''} \n"
    log.error(f"\n{message} Message: {e if e else ''}\n")

def Critical(sector, e, traceback=True):
    _trace = None
    message = ""
    if traceback:
        _trace = trace.format_exc()
        message = f"\n {sector if sector else ''}: {_trace if _trace else ''} \n"
    log.critical(f"\n{message} Message: {e if e else ''}\n")
